'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.defineFunc = defineFunc;
exports.lambda = lambda;
exports.quote = quote;
exports.add = add;
exports.subtract = subtract;
exports.ifFunc = ifFunc;
exports.lteFunc = lteFunc;
exports.display = display;

var _procedure = require('../value/procedure');

var _procedure2 = _interopRequireDefault(_procedure);

var _number = require('../value/number');

var _number2 = _interopRequireDefault(_number);

var _boolean = require('../value/boolean');

var _boolean2 = _interopRequireDefault(_boolean);

var _value = require('../value/value');

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function defineFunc(stack) {
  switch (stack.procTrack) {
    case 0:
      stack.buffer.name = stack.expression.cdr.car;
      if (stack.buffer.name.type !== _value.SYMBOL) {
        throw new Error('Symbol expected, ' + stack.buffer.name.inspect() + ' received');
      }
      this.pushStack(stack.expression.cdr.cdr.car);
      break;
    case 1:
      this.rootParameters[stack.buffer.name.value] = stack.result;
      stack.result = undefined;
      return true;
  }
}

defineFunc.variable = 'define';

function lambda(stack) {
  stack.result = new _procedure2.default('_lambda_', stack.expression.cdr.car, stack.expression.cdr.cdr, stack.scope);
  return true;
}

lambda.variable = 'lambda';

function quote(stack) {
  stack.result = stack.expTrack.car;
  return true;
}

quote.variable = 'quote';

function add(stack) {
  if (stack.procTrack === 0) {
    stack.buffer.sum = 0;
  } else {
    // TODO Type check!
    stack.buffer.sum += stack.result.value;
  }
  if (stack.expTrack) {
    this.pushStack(stack.expTrack.car);
    stack.expTrack = stack.expTrack.cdr;
  } else {
    stack.result = new _number2.default(stack.buffer.sum);
    return true;
  }
}

add.variable = '+';

function subtract(stack) {
  if (stack.procTrack === 1) {
    stack.buffer.sum = stack.result.value;
  } else if (stack.procTrack !== 0) {
    // TODO Type check!
    stack.buffer.sum -= stack.result.value;
  }
  if (stack.expTrack) {
    this.pushStack(stack.expTrack.car);
    stack.expTrack = stack.expTrack.cdr;
  } else {
    stack.result = new _number2.default(stack.buffer.sum);
    return true;
  }
}

subtract.variable = '-';

function ifFunc(stack) {
  switch (stack.procTrack) {
    case 0:
      this.pushStack(stack.expTrack.car);
      stack.expTrack = stack.expTrack.cdr;
      break;
    case 1:
      if (stack.result.type === _value.BOOLEAN && stack.result.value === true) {
        // Follow consequent!
        this.jumpStack(stack.expTrack.car);
        return true;
      } else {
        // Follow alternate!
        if (stack.expTrack.cdr == null) {
          // Unspecified behavior.. What should I do?
          stack.result = undefined;
          return true;
        } else {
          this.jumpStack(stack.expTrack.cdr.car);
          return true;
        }
      }
  }
}

ifFunc.variable = 'if';

// Because I hate WCDMA?
function lteFunc(stack) {
  switch (stack.procTrack) {
    case 0:
      this.pushStack(stack.expTrack.car);
      stack.expTrack = stack.expTrack.cdr;
      break;
    case 1:
      stack.buffer.a = stack.result;
      this.pushStack(stack.expTrack.car);
      break;
    case 2:
      stack.buffer.b = stack.result;
      // Done. :P
      // TODO Type check
      stack.result = new _boolean2.default(stack.buffer.a.value <= stack.buffer.b.value);
      return true;
  }
}

lteFunc.variable = '<=';

function display(stack) {
  if (stack.procTrack === 0) {
    this.pushStack(stack.expTrack.car);
  } else {
    // TODO Display function shouldn't add line break....
    console.log(stack.result.value);
    stack.result = undefined;
    return true;
  }
}

display.variable = 'display';