'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});

var _nativeProcedure = require('../../value/nativeProcedure');

var _nativeProcedure2 = _interopRequireDefault(_nativeProcedure);

var _number = require('../../value/number');

var _number2 = _interopRequireDefault(_number);

var _boolean = require('../../value/boolean');

var _boolean2 = _interopRequireDefault(_boolean);

var _string = require('../../value/string');

var _string2 = _interopRequireDefault(_string);

var _value = require('../../value');

var _createComparator = require('../util/createComparator');

var _createComparator2 = _interopRequireDefault(_createComparator);

var _number3 = require('./number.scm');

var _number4 = _interopRequireDefault(_number3);

var _assert = require('../../util/assert');

var _assert2 = _interopRequireDefault(_assert);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function gcd(a, b) {
  if (a < 0) return gcd(-a, b);
  if (b < 0) return gcd(a, -b);
  if (b > a) return gcd(b, a);
  while (a !== 0 && b !== 0) {
    if (b === 0) return a;
    a %= b;
    if (a === 0) return b;
    b %= a;
  }
}

function lcm(a, b) {
  return a * b / gcd(a, b);
}

function createNumberTest(name, test) {
  return new _nativeProcedure2.default(name, function (list) {
    return new _boolean2.default(list.car && list.car.type === _value.NUMBER && test(list.car.value));
  }, ['obj']);
}

function createNumberCalc(name, callback, single) {
  return new _nativeProcedure2.default(name, function (list) {
    if (list.cdr == null && single) {
      (0, _assert2.default)(list.car, 'number');
      return new _number2.default(single(list.car.value));
    }
    return new _number2.default(list.reduce(function (a, current) {
      var prev = a;
      if (a.type === _value.NUMBER) prev = a.value;
      if (!current || current.type !== _value.NUMBER) return prev;
      return callback(prev, current.value);
    }));
  }, null, 'number');
}

function createNumberOne(name, callback) {
  return new _nativeProcedure2.default(name, function (list) {
    (0, _assert2.default)(list.car, 'number');
    return new _number2.default(callback(list.car.value));
  }, ['a']);
}

function createNumberTwo(name, callback) {
  return new _nativeProcedure2.default(name, function (list) {
    (0, _assert2.default)(list.car, 'number');
    return new _number2.default(callback(list.car.value, list.cdr.car.value));
  }, ['a', 'b']);
}

exports.default = [createNumberTest('number?', function () {
  return true;
}),
// TODO Complex is not supported yet
new _nativeProcedure2.default('complex?', function () {
  return new _boolean2.default(true);
}, []), createNumberTest('real?', function () {
  return true;
}),
// Treat all number as rational number, for now.
createNumberTest('rational?', function () {
  return true;
}), createNumberTest('integer?', function (a) {
  return a % 1 === 0;
}),
// We don't have rational number support yet.. so just check for
// the integers.
createNumberTest('exact?', function (a) {
  return a % 1 === 0;
}), createNumberTest('inexact?', function (a) {
  return a % 1 !== 0;
}), (0, _createComparator2.default)('=', _value.NUMBER, function (a, b) {
  return a === b;
}), (0, _createComparator2.default)('<', _value.NUMBER, function (a, b) {
  return a < b;
}), (0, _createComparator2.default)('>', _value.NUMBER, function (a, b) {
  return a > b;
}), (0, _createComparator2.default)('<=', _value.NUMBER, function (a, b) {
  return a <= b;
}), (0, _createComparator2.default)('>=', _value.NUMBER, function (a, b) {
  return a >= b;
}), createNumberTest('zero?', function (a) {
  return a === 0;
}), createNumberTest('positive?', function (a) {
  return a > 0;
}), createNumberTest('negative?', function (a) {
  return a < 0;
}), createNumberTest('odd?', function (a) {
  return a % 2 === 1;
}), createNumberTest('even?', function (a) {
  return a % 2 === 0;
}), createNumberTest('finite?', function (a) {
  return Number.isFinite(a);
}), createNumberTest('infinite?', function (a) {
  return !Number.isFinite(a);
}), createNumberCalc('max', function (a, b) {
  return Math.max(a, b);
}), createNumberCalc('min', function (a, b) {
  return Math.min(a, b);
}), createNumberCalc('+', function (a, b) {
  return a + b;
}), createNumberCalc('-', function (a, b) {
  return a - b;
}, function (a) {
  return -a;
}), createNumberCalc('/', function (a, b) {
  return a / b;
}, function (a) {
  return 1 / a;
}), createNumberCalc('*', function (a, b) {
  return a * b;
}), createNumberOne('abs', function (a) {
  return Math.abs(a);
}), createNumberCalc('div', function (a, b) {
  return Math.floor(a / b);
}), createNumberCalc('mod', function (a, b) {
  return Math.floor(a % b);
}), createNumberCalc('div0', function (a, b) {
  return a / b | 0;
}), createNumberCalc('mod0', function (a, b) {
  return a % b | 0;
}), createNumberCalc('gcd', function (a, b) {
  return gcd(a, b);
}), createNumberCalc('lcm', function (a, b) {
  return lcm(a, b);
}),
// TODO implement numberator, denominator
createNumberOne('floor', function (a) {
  return Math.floor(a);
}), createNumberOne('ceiling', function (a) {
  return Math.ceiling(a);
}), createNumberOne('truncate', function (a) {
  return a | 0;
}), createNumberOne('round', function (a) {
  return Math.round(a);
}),
// TODO implement rationalize
createNumberOne('exp', function (a) {
  return Math.exp(a);
}), new _nativeProcedure2.default('log', function (list) {
  (0, _assert2.default)(list.car, 'number');
  if (list.cdr == null) {
    return new _number2.default(Math.log(list.car.value));
  } else {
    (0, _assert2.default)(list.cdr, 'number');
    return new _number2.default(Math.log(list.car.value) / Math.log(list.cdr.car.value));
  }
}, ['a', 'b']), createNumberOne('sin', function (a) {
  return Math.sin(a);
}), createNumberOne('cos', function (a) {
  return Math.cos(a);
}), createNumberOne('tan', function (a) {
  return Math.tan(a);
}), createNumberOne('asin', function (a) {
  return Math.asin(a);
}), createNumberOne('acos', function (a) {
  return Math.acos(a);
}), new _nativeProcedure2.default('atan', function (list) {
  (0, _assert2.default)(list.car, 'number');
  if (list.cdr == null) {
    return new _number2.default(Math.atan(list.car.value));
  } else {
    (0, _assert2.default)(list.cdr, 'number');
    return new _number2.default(Math.atan2(list.car.value, list.cdr.car.value));
  }
}, ['a', 'b']), createNumberOne('sqrt', function (a) {
  return Math.sqrt(a);
}),
// TODO exact-integer-sqrt
createNumberTwo('expt', function (a, b) {
  return Math.pow(a, b);
}),
// TODO make-rectangular, make-polar, real-part, imag-part, magnitude, angle
// these require complex numbers (which is going to be quite complex :P)
new _nativeProcedure2.default('number->string', function (list) {
  // Ignore precision for now
  (0, _assert2.default)(list.car, 'number');
  if (list.car.value === Infinity) return new _string2.default('inf.0');
  if (list.car.value === -Infinity) return new _string2.default('-inf.0');
  if (isNaN(list.car.value)) return new _string2.default('nan.0');
  return new _string2.default(list.car.value.toString(list.cdr.car.value));
}, ['number']), new _nativeProcedure2.default('string->number', function (list) {
  (0, _assert2.default)(list.car, 'string');
  if (list.car.value === 'inf.0' || list.car.value === '+inf.0') {
    return new _number2.default(Infinity);
  }
  if (list.car.value === '-inf.0') {
    return new _number2.default(-Infinity);
  }
  if (/^[\-+]?nan.0$/.test(list.car.value)) {
    return new _number2.default(NaN);
  }
  return new _number2.default(parseFloat(list.car.value, list.cdr.car.value));
}, ['string']), _number4.default];