'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});

var _nativeProcedure = require('../../value/nativeProcedure');

var _nativeProcedure2 = _interopRequireDefault(_nativeProcedure);

var _boolean = require('../../value/boolean');

var _boolean2 = _interopRequireDefault(_boolean);

var _pair = require('../../value/pair');

var _pair2 = _interopRequireDefault(_pair);

var _number = require('../../value/number');

var _number2 = _interopRequireDefault(_number);

var _value = require('../../value');

var _pair3 = require('./pair.scm');

var _pair4 = _interopRequireDefault(_pair3);

var _assert = require('../../util/assert');

var _assert2 = _interopRequireDefault(_assert);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

exports.default = [new _nativeProcedure2.default('pair?', function (list) {
  return new _boolean2.default(list.car && list.car.type === _value.PAIR);
}, ['obj']), new _nativeProcedure2.default('null?', function (list) {
  if (list.car == null) return _boolean.TRUE;
  if (list.car.type !== _value.PAIR) return _boolean.FALSE;
  return new _boolean2.default(list.car.isEmpty());
}, ['obj']), new _nativeProcedure2.default('list?', function (list) {
  return new _boolean2.default(list.car && list.car.type === _value.PAIR && list.car.isList());
}, ['obj']), new _nativeProcedure2.default('cons', function (list) {
  return new _pair2.default(list.car, list.cdr.car);
}, ['car', 'cdr']), new _nativeProcedure2.default('car', function (list) {
  if (list.car.car == null) {
    throw new Error('Empty pair or non-pair received');
  }
  return list.car.car;
}, ['pair']), new _nativeProcedure2.default('cdr', function (list) {
  if (list.car.car == null) {
    throw new Error('Empty pair or non-pair received');
  }
  return list.car.cdr || new _pair2.default();
}, ['pair']), new _nativeProcedure2.default('list', function (list) {
  return list;
}, null, 'obj'), new _nativeProcedure2.default('length', function (list) {
  (0, _assert2.default)(list.car, 'pair');
  return new _number2.default(list.car.length());
}, ['list']), new _nativeProcedure2.default('append', function (list) {
  var head = void 0,
      tail = void 0;
  var outerNode = list;
  while (outerNode != null && outerNode.type === _value.PAIR) {
    var node = outerNode.car;
    outerNode = outerNode.cdr;
    while (node != null && node.type === _value.PAIR) {
      var o = new _pair2.default(node.car, null);
      if (tail) {
        tail.cdr = o;
        tail = o;
      } else {
        head = o;
        tail = o;
      }
      node = node.cdr;
    }
    if (node != null) {
      if (outerNode == null) {
        if (tail) {
          tail.cdr = node;
        } else {
          // If no pairs are allocated, obj is returned.
          return node;
        }
      } else {
        var _o = new _pair2.default(node, null);
        if (tail) {
          tail.cdr = _o;
          tail = _o;
        } else {
          head = _o;
          tail = _o;
        }
      }
    }
  }
  return head || new _pair2.default();
}, null, 'obj'), new _nativeProcedure2.default('reverse', function (list) {
  (0, _assert2.default)(list.car, 'pair');
  var output = void 0;
  var node = list.car;
  while (node != null && node.type === _value.PAIR) {
    output = new _pair2.default(node.car, output);
    node = node.cdr;
  }
  return output || new _pair2.default();
}, ['list']), new _nativeProcedure2.default('map', function (list, machine, frame) {
  var procedure = list.car;
  (0, _assert2.default)(procedure, 'procedure');
  var argsHeader = new _pair2.default(procedure);
  var argsTail = argsHeader;

  if (frame.bufferHead == null) {
    frame.bufferHead = list.cdr.copy();
    frame.result = null;
  }
  if (frame.result != null) {
    var pair = new _pair2.default(frame.result);
    if (frame.bufferOutputHead == null) {
      frame.bufferOutputHead = pair;
      frame.bufferOutputTail = pair;
    } else {
      frame.bufferOutputTail.cdr = pair;
      frame.bufferOutputTail = pair;
    }
    frame.result = null;
  }
  var headHead = frame.bufferHead;
  var headNode = null;
  if (headHead.car != null) {
    headNode = headHead;
    // Visit every node, and increment each node.
    while (headNode != null && headNode.type === _value.PAIR) {
      if (headNode.car == null || headNode.car.type !== _value.PAIR) {
        throw new Error('List sizes does not match');
      }
      var argsPair = new _pair2.default(headNode.car.car);
      argsTail.cdr = argsPair;
      argsTail = argsPair;

      headNode.car = headNode.car.cdr;
      headNode = headNode.cdr;
    }

    // Push the stack frame...
    machine.pushStack(argsHeader, null, true);
    return undefined;
  }
  // Done!
  return frame.bufferOutputHead;
}, ['proc'], 'list'), new _nativeProcedure2.default('for-each', function (list, machine, frame) {
  // R6RS standard requires that more than one list should be handled.
  // This can be implemented kinda absurdly....
  var procedure = list.car;
  (0, _assert2.default)(procedure, 'procedure');
  var argsHeader = new _pair2.default(procedure);
  var argsTail = argsHeader;

  if (frame.bufferHead == null) frame.bufferHead = list.cdr.copy();
  var headHead = frame.bufferHead;
  var headNode = null;
  if (headHead.car != null) {
    headNode = headHead;
    // Visit every node, and increment each node.
    while (headNode != null && headNode.type === _value.PAIR) {
      if (headNode.car == null || headNode.car.type !== _value.PAIR) {
        throw new Error('List sizes does not match');
      }
      var argsPair = new _pair2.default(headNode.car.car);
      argsTail.cdr = argsPair;
      argsTail = argsPair;

      headNode.car = headNode.car.cdr;
      headNode = headNode.cdr;
    }

    // Push the stack frame...
    machine.pushStack(argsHeader, null, true);
    return undefined;
  }
  // Done!
  return new _pair2.default();
}, ['proc'], 'list'), _pair4.default];