'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});

var _lambda = require('../../value/lambda');

var _lambda2 = _interopRequireDefault(_lambda);

var _nativeProcedure = require('../../value/nativeProcedure');

var _nativeProcedure2 = _interopRequireDefault(_nativeProcedure);

var _nativeSyntax = require('../../value/nativeSyntax');

var _nativeSyntax2 = _interopRequireDefault(_nativeSyntax);

var _boolean = require('../../value/boolean');

var _boolean2 = _interopRequireDefault(_boolean);

var _pair2 = require('../../value/pair');

var _pair3 = _interopRequireDefault(_pair2);

var _value = require('../../value');

var _primitive = require('./primitive.scm');

var _primitive2 = _interopRequireDefault(_primitive);

var _pair4 = require('./pair');

var _pair5 = _interopRequireDefault(_pair4);

var _boolean3 = require('./boolean');

var _boolean4 = _interopRequireDefault(_boolean3);

var _symbol = require('./symbol');

var _symbol2 = _interopRequireDefault(_symbol);

var _character = require('./character');

var _character2 = _interopRequireDefault(_character);

var _string = require('./string');

var _string2 = _interopRequireDefault(_string);

var _number = require('./number');

var _number2 = _interopRequireDefault(_number);

var _assert = require('../../util/assert');

var _assert2 = _interopRequireDefault(_assert);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

// Base library - Primitive functions that depend on native calls.

exports.default = [
// Other forms will be handled by define-syntax. :)
new _nativeSyntax2.default('define', function (machine, frame) {
  switch (frame.procTrack) {
    case 0:
      frame.bufferName = frame.expTrack.car;
      if (frame.bufferName.type !== _value.SYMBOL) {
        throw new Error('Symbol expected, ' + frame.bufferName.inspect() + ' received');
      }
      machine.pushStack(frame.expTrack.cdr.car);
      break;
    case 1:
      if (machine.libraryLevel) {
        machine.libraryParameters[frame.bufferName.value] = frame.result;
      } else {
        machine.rootParameters[frame.bufferName.value] = frame.result;
      }
      frame.result = new _pair3.default();
      return true;
  }
}, ['variable', 'expression']), new _nativeSyntax2.default('set!', function (machine, frame) {
  switch (frame.procTrack) {
    case 0:
      frame.bufferName = frame.expTrack.car;
      if (frame.bufferName.type !== _value.SYMBOL) {
        throw new Error('Symbol expected, ' + frame.bufferName.inspect() + ' received');
      }
      machine.pushStack(frame.expTrack.cdr.car);
      break;
    case 1:
      {
        var name = frame.bufferName.value;
        var value = frame.result;
        frame.result = new _pair3.default();
        var node = frame.scope;
        while (node != null) {
          if (node.car[name] != null) {
            node.car[name] = value;
            return true;
          }
          node = node.cdr;
        }
        if (machine.rootParameters[name] != null) {
          machine.rootParameters[name] = value;
        }
        if (machine.libraryParameters[name] != null) {
          if (machine.libraryLevel) {
            machine.libraryParameters[name] = value;
          } else {
            machine.rootParameters[name] = value;
          }
        }
        return true;
      }}
}, ['variable', 'expression']),
// define-syntax is processed by expander, so machine itself doesn't have to
// process it at all.
new _nativeSyntax2.default('define-syntax', function () {
  // NOP
  return true;
}, ['keyword', 'expression']), new _nativeSyntax2.default('let-syntax', function () {
  // NOP
  return true;
}, ['keyword', 'expression']), new _nativeSyntax2.default('letrec-syntax', function () {
  // NOP
  return true;
}, ['keyword', 'expression']), new _nativeSyntax2.default('lambda', function (machine, frame) {
  // If parent node is define or set! (at library level), we can pull the
  // name from previous stack and use that instead.
  var name = '_lambda_',
      nameValue = void 0;
  if (machine.stack.cdr && machine.stack.cdr.car) {
    var prevStack = machine.stack.cdr.car;
    if (prevStack.procedure === machine.libraryParameters['define'] || prevStack.procedure === machine.libraryParameters['set!']) {
      if (prevStack.expTrack.car && prevStack.expTrack.car.type === _value.SYMBOL) {
        name = prevStack.expTrack.car.value;
        nameValue = prevStack.expTrack.car;
      }
    }
  }
  frame.result = new _lambda2.default(name, frame.expTrack.cdr, frame.expTrack.car, frame.scope);
  frame.result.nameVal = nameValue;
  return true;
}, ['formals'], 'body'), new _nativeSyntax2.default('quote', function (machine, frame) {
  frame.result = frame.expTrack.car;
  return true;
}, ['datum']), new _nativeSyntax2.default('quasiquote', function (machine, frame) {
  if (frame.bufferStack == null) {
    // Initialize quasiquote traverser
    frame.bufferStack = new _pair3.default({
      input: frame.expTrack.car
    });
  }
  function pushValue(entry, value) {
    var pair = new _pair3.default(value);
    if (entry.outputTail) {
      entry.outputTail.cdr = pair;
      entry.outputTail = pair;
    } else {
      entry.outputHead = pair;
      entry.outputTail = pair;
    }
  }
  while (frame.bufferStack != null) {
    var entry = frame.bufferStack.car;
    if (entry.input == null) {
      frame.bufferStack = frame.bufferStack.cdr;
      if (frame.bufferStack != null) {
        frame.bufferStack.car.result = entry.outputHead || new _pair3.default();
      } else {
        frame.result = entry.outputHead || new _pair3.default();
        return true;
      }
    } else if (entry.input.type === _value.PAIR) {
      if (entry.input.car.type === _value.PAIR) {
        var subCode = entry.input.car;
        if (subCode.car.type === _value.SYMBOL && subCode.car.value === 'unquote') {
          if (frame.result != null) {
            pushValue(entry, frame.result);
            frame.result = null;
          } else {
            machine.pushStack(subCode.cdr.car);
            return;
          }
        } else if (subCode.car.type === _value.SYMBOL && subCode.car.value === 'unquote-splicing') {
          if (frame.result != null) {
            var listNode = frame.result;
            while (listNode != null && listNode.type === _value.PAIR) {
              pushValue(entry, listNode.car);
              listNode = listNode.cdr;
            }
            frame.result = null;
          } else {
            machine.pushStack(subCode.cdr.car);
            return;
          }
        } else {
          if (entry.result) {
            pushValue(entry, entry.result);
            entry.result = null;
          } else {
            frame.bufferStack = new _pair3.default({
              input: subCode
            }, frame.bufferStack);
            continue;
          }
        }
      } else {
        pushValue(entry, entry.input.car);
      }
      entry.input = entry.input.cdr;
    } else {
      // cdr value.
      if (entry.outputTail) {
        entry.outputTail.cdr = entry.input;
      } else {
        var _pair = new _pair3.default(null, entry.input);
        entry.outputHead = entry.outputTail = _pair;
      }
      entry.input = null;
    }
  }
  frame.result = frame.expTrack.car;
  return true;
}, ['datum']), new _nativeSyntax2.default('if', function (machine, frame) {
  switch (frame.procTrack) {
    case 0:
      machine.pushStack(frame.expTrack.car);
      frame.expTrack = frame.expTrack.cdr;
      break;
    case 1:
      if (frame.result.type !== _value.BOOLEAN || frame.result.value === true) {
        // Follow consequent!
        machine.jumpStack(frame.expTrack.car);
        return true;
      } else {
        // Follow alternate!
        if (frame.expTrack.cdr == null) {
          // Unspecified behavior.. What should I do?
          frame.result = new _boolean2.default(false);
          return true;
        } else {
          machine.jumpStack(frame.expTrack.cdr.car);
          return true;
        }
      }
  }
}, ['test', 'consequent', 'alternate']), new _nativeProcedure2.default('begin', function (list) {
  var result = null;
  var node = list;
  while (node != null) {
    result = node.car;
    node = node.cdr;
  }
  return result;
}, null, 'form'), new _nativeProcedure2.default('apply', function (list, machine) {
  (0, _assert2.default)(list.car, _value.PROCEDURE);
  var listHead = new _pair3.default(list.car);
  var listTail = listHead;
  var node = list.cdr;
  while (node != null && node.cdr != null) {
    var pairValue = new _pair3.default(node.car);
    listTail.cdr = pairValue;
    listTail = pairValue;
    node = node.cdr;
  }
  (0, _assert2.default)(node.car, _value.PAIR);
  listTail.cdr = node.car;
  machine.jumpStack(listHead, true);
  return true;
}, ['proc'], 'args'), new _nativeProcedure2.default('error', function (list) {
  var output = 'Error: ';
  if (list.car && list.car.value !== false) {
    output += list.car.value + ': ';
  }
  output += list.cdr.car.value + ' ';
  output += list.cdr.cdr.inspect();
  throw new Error(output);
}, ['who', 'message'], 'irritant'), new _nativeProcedure2.default('assertion-violation', function (list) {
  var output = 'Assertion violation: ';
  if (list.car && list.car.value !== false) {
    output += list.car.value + ': ';
  }
  output += list.cdr.car.value + ' ';
  output += list.cdr.cdr.inspect();
  throw new Error(output);
}, ['who', 'message'], 'irritant'), new _nativeProcedure2.default('assert', function (list, _, frame) {
  var val = list.car;
  if (val && val.type === _value.BOOLEAN && val.value === false) {
    // Assert if false
    throw new Error('Assertion violation: ' + frame.expression.cdr.inspect());
  }
  // Return if true
  return val;
}, ['expression']), new _nativeProcedure2.default('eqv?', function (list) {
  var a = list.car;
  var b = list.cdr.car;
  if (a === b) return _boolean2.default.TRUE;
  if (a.type !== b.type) return _boolean2.default.FALSE;
  switch (a.type) {
    case _value.BOOLEAN:
    case _value.SYMBOL:
    case _value.NUMBER:
    case _value.CHARACTER:
      return new _boolean2.default(a.value === b.value);
    case _value.PAIR:
      if (a.isEmpty() && b.isEmpty()) return _boolean2.default.TRUE;
      return new _boolean2.default(a === b);
    default:
      return new _boolean2.default(a === b);
  }
}, ['obj1', 'obj2']), function () {
  function checkEqual(a, b) {
    if (a == b) return true;
    if (a == null || b == null) return false;
    if (a.type !== b.type) return false;
    switch (a.type) {
      case _value.BOOLEAN:
      case _value.SYMBOL:
      case _value.NUMBER:
      case _value.CHARACTER:
      case _value.STRING:
        return a.value === b.value;
      case _value.PAIR:
        if (!checkEqual(a.car, b.car)) return false;
        if (!checkEqual(a.cdr, b.cdr)) return false;
        return true;
      default:
        return a === b;
    }
  }
  return new _nativeProcedure2.default('equal?', function (list) {
    var a = list.car;
    var b = list.cdr.car;
    return new _boolean2.default(checkEqual(a, b));
  }, ['obj1', 'obj2']);
}(), new _nativeProcedure2.default('procedure?', function (list) {
  return new _boolean2.default(list.car && list.car.type === _value.PROCEDURE);
}, ['obj']), _primitive2.default, _pair5.default, _boolean4.default, _symbol2.default, _character2.default, _string2.default, _number2.default];