'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});

var _nativeProcedure = require('../../value/nativeProcedure');

var _nativeProcedure2 = _interopRequireDefault(_nativeProcedure);

var _character = require('../../value/character');

var _character2 = _interopRequireDefault(_character);

var _string = require('../../value/string');

var _string2 = _interopRequireDefault(_string);

var _number = require('../../value/number');

var _number2 = _interopRequireDefault(_number);

var _boolean = require('../../value/boolean');

var _boolean2 = _interopRequireDefault(_boolean);

var _symbol = require('../../value/symbol');

var _symbol2 = _interopRequireDefault(_symbol);

var _pair = require('../../value/pair');

var _pair2 = _interopRequireDefault(_pair);

var _value = require('../../value');

var _assert = require('../../util/assert');

var _assert2 = _interopRequireDefault(_assert);

var _createComparator = require('../util/createComparator');

var _createComparator2 = _interopRequireDefault(_createComparator);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function stringToList(string) {
  var head = void 0,
      tail = void 0;
  // Assert string
  for (var i = 0; i < string.length; ++i) {
    var pair = new _pair2.default(new _character2.default(string[i]));
    if (head == null) {
      head = pair;
      tail = pair;
    } else {
      tail.cdr = pair;
      tail = pair;
    }
  }
  return head || new _pair2.default();
}

exports.default = [new _nativeProcedure2.default('string?', function (list) {
  return new _boolean2.default(list.car && list.car.type === _value.STRING);
}, ['obj']), new _nativeProcedure2.default('make-string', function (list) {
  (0, _assert2.default)(list.car, 'number');
  (0, _assert2.default)(list.cdr.car, 'character');
  var letters = list.car.value;
  var char = list.cdr && list.cdr.car.value || '\0';
  var output = '';
  for (var i = 0; i < letters; ++i) {
    output += char;
  }
  return new _string2.default(output);
}, ['k', 'char']), new _nativeProcedure2.default('string', function (list) {
  var output = '';
  list.forEach(function (char) {
    (0, _assert2.default)(char, 'character');
    output += char.value;
  });
  return new _string2.default(output);
}, null, 'char'), new _nativeProcedure2.default('string-length', function (list) {
  (0, _assert2.default)(list.car, 'string');
  return new _number2.default(list.car.value.length);
}, ['string']), new _nativeProcedure2.default('string-ref', function (list) {
  // Assert string, integer, range
  (0, _assert2.default)(list.car, 'string');
  (0, _assert2.default)(list.cdr.car, 'number');
  return new _character2.default(list.car.value[list.cdr.car.value]);
}, ['string', 'k']), (0, _createComparator2.default)('string=?', _value.STRING, function (a, b) {
  return a === b;
}), (0, _createComparator2.default)('string<?', _value.STRING, function (a, b) {
  return a < b;
}), (0, _createComparator2.default)('string>?', _value.STRING, function (a, b) {
  return a > b;
}), (0, _createComparator2.default)('string<=?', _value.STRING, function (a, b) {
  return a <= b;
}), (0, _createComparator2.default)('string>=?', _value.STRING, function (a, b) {
  return a >= b;
}), new _nativeProcedure2.default('substring', function (list) {
  // Assert string, integer, integer
  (0, _assert2.default)(list.car, 'string');
  (0, _assert2.default)(list.cdr.car, 'number');
  (0, _assert2.default)(list.cdr.cdr.car, 'number');
  var original = list.car.value;
  var start = list.cdr.car.value;
  var end = list.cdr.cdr.car.value;
  return new _string2.default(original.slice(start, end));
}, ['string', 'start', 'end']), new _nativeProcedure2.default('string-append', function (list) {
  var output = '';
  list.forEach(function (string) {
    // Assert string
    (0, _assert2.default)(string, 'string');
    output += string.value;
  });
  return new _string2.default(output);
}, null, 'string'), new _nativeProcedure2.default('string->list', function (list) {
  (0, _assert2.default)(list.car, 'string');
  return stringToList(list.car.value);
}, ['string']), new _nativeProcedure2.default('list->string', function (list) {
  var output = '';
  (0, _assert2.default)(list.car, 'pair');
  list.car.forEach(function (char) {
    (0, _assert2.default)(char, 'character');
    output += char.value;
  });
  return new _string2.default(output);
}, ['list']), new _nativeProcedure2.default('string-for-each', function (list, machine, frame) {
  // All this does is convert the request to for-each request. Easy!
  if (frame.procTrack === 0) {
    var request = new _pair2.default(new _symbol2.default('for-each'), new _pair2.default(list.car, list.cdr.map(function (v) {
      (0, _assert2.default)(v, 'string');
      return new _pair2.default(new _symbol2.default('quote'), new _pair2.default(stringToList(v.value)));
    })));
    machine.pushStack(request);
  } else {
    return frame.result;
  }
}, ['proc'], 'string'), new _nativeProcedure2.default('string-copy', function (list) {
  return new _string2.default(list.car.value);
}, ['string'])];