'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});

var _createClass = function () { function defineProperties(target, props) { for (var i = 0; i < props.length; i++) { var descriptor = props[i]; descriptor.enumerable = descriptor.enumerable || false; descriptor.configurable = true; if ("value" in descriptor) descriptor.writable = true; Object.defineProperty(target, descriptor.key, descriptor); } } return function (Constructor, protoProps, staticProps) { if (protoProps) defineProperties(Constructor.prototype, protoProps); if (staticProps) defineProperties(Constructor, staticProps); return Constructor; }; }(); // A object representing LISP machine.


var _value = require('./value');

var _pair = require('./value/pair');

var _pair2 = _interopRequireDefault(_pair);

var _parser = require('./parser');

var _parser2 = _interopRequireDefault(_parser);

var _tokenizer = require('./tokenizer');

var _tokenizer2 = _interopRequireDefault(_tokenizer);

var _expander = require('./expander');

var _expander2 = _interopRequireDefault(_expander);

var _function = require('./function');

var _function2 = _interopRequireDefault(_function);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _classCallCheck(instance, Constructor) { if (!(instance instanceof Constructor)) { throw new TypeError("Cannot call a class as a function"); } }

var Machine = function () {
  function Machine() {
    var loadBase = arguments.length <= 0 || arguments[0] === undefined ? true : arguments[0];
    var libraryCache = arguments[1];

    _classCallCheck(this, Machine);

    // Stores root parameter information - user defined variables, functions,
    // etc.
    this.rootParameters = {};
    // Root syntax information used by the expander.
    this.expanderRoot = {};

    if (libraryCache != null) {
      if (libraryCache.parameters == null) {
        libraryCache.parameters = {};
      }
      if (libraryCache.expands == null) {
        libraryCache.expands = {};
      }
      this.libraryParameters = libraryCache.parameters;
      this.expanderLibrary = libraryCache.expands;
    } else {
      // Stores library parameter information - base library, SRFI, etc.
      this.libraryParameters = {};
      // Library root syntax information used by the expander.
      this.expanderLibrary = {};
    }

    // The execute stack. It uses cons (PairValue) to store information.
    // Note that this only stores 'execute' stack; scoped variables are stored
    // in different place.
    this.stack = null;
    this.stackDepth = 0;
    this.libraryLevel = false;

    // The code quota. Code execution will stop after reaching the threshold.
    // (Used to prevent infinite loop hell)
    this.quota = 0;

    // The standard I/O. Only stdout is available at this moment though.
    // Since ports are not supported yet, the stdout function should be a
    // function that accepts a string.
    if (typeof process !== 'undefined') {
      this.stdout = function (str) {
        return process.stdout.write(str);
      };
    } else {
      // Fallback to console.log in browsers (Which is awful)
      this.stdout = function (str) {
        return console.log(str);
      };
    }

    if (loadBase) this.loadLibrary(_function2.default);
  }

  _createClass(Machine, [{
    key: 'getVariable',
    value: function getVariable(name) {
      // Iterate until scope appears...
      var node = this.stack && this.stack.car.scope;
      while (node != null) {
        if (node.car[name] != null) return node.car[name];
        node = node.cdr;
      }
      if (this.rootParameters[name] != null) {
        return this.rootParameters[name];
      }
      if (this.libraryParameters[name] != null) {
        return this.libraryParameters[name];
      }
      throw new Error('Unbound variable: ' + name);
    }
  }, {
    key: 'jumpStack',
    value: function jumpStack(list, noResolve) {
      // Performs TCO optimization
      var stackEntry = this.stack.car;
      this.popStack();
      this.pushStack(list, stackEntry, noResolve);
      stackEntry.tco = true;
    }
  }, {
    key: 'clearStack',
    value: function clearStack() {
      this.stack = null;
      this.stackDepth = 0;
    }
  }, {
    key: 'popStack',
    value: function popStack() {
      this.stack = this.stack.cdr;
      this.stackDepth--;
    }
  }, {
    key: 'pushStack',
    value: function pushStack(list, stackEntry, noResolve) {
      var scope = void 0;
      if (stackEntry) {
        scope = stackEntry.scope;
      } else {
        scope = this.stack && this.stack.car.scope;
      }
      this.stack = new _pair2.default({
        expression: list,
        scope: scope, noResolve: noResolve
      }, this.stack);
      this.stackDepth++;
    }
    // Returns the stacktrace if the machine has abnormally exit.

  }, {
    key: 'getStackTrace',
    value: function getStackTrace() {
      var fancy = arguments.length <= 0 || arguments[0] === undefined ? false : arguments[0];

      var result = [];
      var entry = this.stack;
      while (entry != null) {
        var stackEntry = entry.car;
        if (stackEntry.expression && stackEntry.expression.car) {
          var name = stackEntry.expression.car;
          if (name.line != null) {
            result.push(name.value + ' (' + name.line + ':' + name.column + ')');
          } else {
            result.push(name.value);
          }
        } else {
          // ???
          result.push('<unknown>');
        }
        entry = entry.cdr;
      }
      if (fancy) {
        return result.map(function (v) {
          return '    at ' + v;
        }).join('\n');
      }
      return result.join('\n');
    }
  }, {
    key: 'execute',
    value: function execute() {
      var startStackDepth = this.stackDepth;
      var ops = 0;
      // Loop until the stack ends...
      while (this.stackDepth >= startStackDepth && this.stack != null) {
        if (this.stackDepth >= 65536) {
          throw new Error('Stack overflow');
        }
        if (this.quota > 0 && ops >= this.quota) {
          // Quota exceeded
          throw new Error('Quota exceeded');
        }
        ops++;
        var stackData = this.stack.car;
        var expression = stackData.expression;
        var procedure = stackData.procedure;
        var result = stackData.result;

        if (expression.type !== _value.PAIR) {
          // If constant values are provided...
          var _runResult = void 0;
          if (expression.type === _value.SYMBOL) {
            _runResult = this.getVariable(expression.value);
          } else {
            _runResult = expression;
          }
          // Code is complete; Return the value and release stack.
          this.popStack();
          if (this.stackDepth >= startStackDepth && this.stack) {
            // Set the result value of the stack.
            this.stack.car.result = _runResult;
            continue;
          } else {
            // If no entry is available, just return the result.
            return _runResult;
          }
        }
        if (procedure == null) {
          // Procedure is not ready yet; Is the procedure calculated?
          if (result == null) {
            // Nope! Try to resolve the procedure. If value is a symbol,
            // resolve it without creating new stack entry.
            // If value is a procedure (It's not possible...), solve it directly.
            // If value is a list (pair), create new stack entry.
            var original = expression.car;
            if (original == null) {
              throw new Error('List expected; got empty pair instead');
            }
            if (original.type === _value.SYMBOL) {
              procedure = this.getVariable(original.value);
              stackData.procedure = procedure;
            } else if (original.type === _value.PROCEDURE) {
              procedure = original;
              stackData.procedure = procedure;
            } else if (original.type === _value.PAIR) {
              // Create new stack entry and run that instead.
              this.pushStack(original);
              continue;
            } else {
              // Raise an exception.
              throw new Error('Procedure expected, got ' + original.inspect() + ' instead');
            }
          } else {
            // If procedure is calculated, just continue.
            procedure = result;
            stackData.procedure = result;
          }
          // We have to check validity of the procedure - It should be a
          // procedure object.
          if (procedure.type !== _value.PROCEDURE) {
            // Raise an exception; However since we lack stack rewinding and
            // stuff (such as with-exception-handler), just throw an native
            // exception.
            throw new Error('Procedure expected, got ' + procedure.inspect() + ' instead');
          }
        }
        // We've got the procedure - Let's pass the whole stack frame to
        // the procedure!
        var runResult = procedure.execute(this, stackData);
        // True indicates that the executing is over.
        if (runResult === true && !stackData.tco) {
          result = stackData.result;
          // Code is complete; Return the value and release stack.
          this.popStack();
          if (this.stackDepth >= startStackDepth && this.stack) {
            // Set the result value of the stack.
            this.stack.car.result = result;
            continue;
          } else {
            // If no entry is available, just return the result.
            return result;
          }
        }
      }
    }
    // This accepts AST generated by parser; it can't process raw string!
    // Direct means that the provided code should be treated as single list,
    // thus preventing separation.
    // Library level means the evaluation is running in the library level, thus
    // it should define variables in library scope.

  }, {
    key: 'evaluate',
    value: function evaluate(code) {
      var direct = arguments.length <= 1 || arguments[1] === undefined ? false : arguments[1];
      var libraryLevel = arguments.length <= 2 || arguments[2] === undefined ? false : arguments[2];

      var ast = void 0;
      if (typeof code === 'string') {
        if (libraryLevel) {
          ast = (0, _expander2.default)((0, _parser2.default)((0, _tokenizer2.default)(code)), this.expanderLibrary);
        } else {
          ast = (0, _expander2.default)((0, _parser2.default)((0, _tokenizer2.default)(code)), this.expanderRoot, new _pair2.default(this.expanderLibrary));
        }
      } else {
        ast = code;
      }
      if (ast == null) return new _pair2.default();
      this.libraryLevel = libraryLevel;
      if (direct || ast.type !== _value.PAIR) {
        this.pushStack(ast);
        return this.execute();
      } else {
        // Process one by one....
        var node = ast;
        var result = void 0;
        if (node.car == null) return new _pair2.default();
        while (node !== null && node.type === _value.PAIR) {
          this.pushStack(node.car);
          result = this.execute();
          node = node.cdr;
        }
        // Should we process cdr value too?
        if (node !== null) {
          this.pushStack(node);
          result = this.execute();
        }
        return result;
      }
    }
    // Loads platform library into the interpreter scope.

  }, {
    key: 'loadLibrary',
    value: function loadLibrary(list) {
      var _iteratorNormalCompletion = true;
      var _didIteratorError = false;
      var _iteratorError = undefined;

      try {
        for (var _iterator = list[Symbol.iterator](), _step; !(_iteratorNormalCompletion = (_step = _iterator.next()).done); _iteratorNormalCompletion = true) {
          var entry = _step.value;

          if (entry.type === _value.PROCEDURE) {
            this.libraryParameters[entry.name] = entry;
          } else if (Array.isArray(entry)) {
            this.loadLibrary(entry);
          } else if (typeof entry === 'string') {
            this.evaluate(entry, false, true);
          }
        }
      } catch (err) {
        _didIteratorError = true;
        _iteratorError = err;
      } finally {
        try {
          if (!_iteratorNormalCompletion && _iterator.return) {
            _iterator.return();
          }
        } finally {
          if (_didIteratorError) {
            throw _iteratorError;
          }
        }
      }
    }
  }]);

  return Machine;
}();

exports.default = Machine;