'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});

var _createClass = function () { function defineProperties(target, props) { for (var i = 0; i < props.length; i++) { var descriptor = props[i]; descriptor.enumerable = descriptor.enumerable || false; descriptor.configurable = true; if ("value" in descriptor) descriptor.writable = true; Object.defineProperty(target, descriptor.key, descriptor); } } return function (Constructor, protoProps, staticProps) { if (protoProps) defineProperties(Constructor.prototype, protoProps); if (staticProps) defineProperties(Constructor, staticProps); return Constructor; }; }();

var _value = require('../value');

var _pair2 = require('../value/pair');

var _pair3 = _interopRequireDefault(_pair2);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _classCallCheck(instance, Constructor) { if (!(instance instanceof Constructor)) { throw new TypeError("Cannot call a class as a function"); } }

var LIST_WRAP = Symbol('listWrap');

var SyntaxRules = function () {
  function SyntaxRules(list) {
    var _this = this;

    _classCallCheck(this, SyntaxRules);

    this.literals = [];
    list.cdr.car.forEach(function (literal) {
      if (literal.type !== _value.SYMBOL) throw new Error('Literal must be a symbol');
      _this.literals.push(literal);
    });
    this.rules = list.cdr.cdr;
  }

  _createClass(SyntaxRules, [{
    key: 'checkPattern',
    value: function checkPattern(pattern, code, scope) {
      var parentEllipsis = arguments.length <= 3 || arguments[3] === undefined ? false : arguments[3];

      var patternLen = pattern ? pattern.length(false) : 0;
      var codeLen = code ? code.length(false) : 0;
      var patternCur = pattern;
      var patternNext = pattern.cdr;
      var patternEllipsis = patternNext && patternNext.type === _value.PAIR && patternNext.car.type === _value.SYMBOL && patternNext.car.value === '...';
      var codeNode = code;
      while (patternCur != null && patternCur.type === _value.PAIR) {
        if (codeNode == null && (patternEllipsis || parentEllipsis)) {
          if (patternCur.car.type === _value.SYMBOL && scope[patternCur.car.value] == null) {
            // Create empty value to prevent template error
            scope[patternCur.car.value] = {
              type: LIST_WRAP
            };
          } else if (patternCur.car.type === _value.PAIR) {
            // Create empty value (recursion)
            this.checkPattern(patternCur.car, null, scope, patternEllipsis);
          }
          patternCur = patternNext;
          patternNext = patternNext && patternNext.cdr;
          patternEllipsis = patternNext && patternNext.type === _value.PAIR && patternNext.car.type === _value.SYMBOL && patternNext.car.value === '...';
          patternLen--;
          continue;
        }
        if (patternCur.car == null) {
          // Do absolutely nothing
          if (codeNode != null && codeNode.car != null) {
            // Null check fail
            return false;
          }
        } else if (patternCur.car.type === _value.SYMBOL) {
          if (patternCur.car.value === '...') {
            // This should just be skipped, without triggering code change
            patternCur = patternNext;
            patternNext = patternNext && patternNext.cdr;
            patternEllipsis = patternNext && patternNext.type === _value.PAIR && patternNext.car.type === _value.SYMBOL && patternNext.car.value === '...';
            patternLen--;
            continue;
          } else if (codeNode == null) {
            // If codeNode is null - it failed (underflow).
            if (codeNode == null) return false;
          } else if (patternCur.car.value === '_') {
            // This always matches
          } else if (this.literals.indexOf(patternCur.car.value) !== -1) {
              // Literal identifier; check if code also matches.
              if (codeNode.car.type !== _value.SYMBOL || patternCur.car.value !== codeNode.car.value) return false;
            } else {
              // Pattern variable...
              // TODO check confliction
              if (!parentEllipsis && !patternEllipsis) {
                // Flat pattern. We can just copy it to the scope
                scope[patternCur.car.value] = codeNode.car;
              } else {
                // Ellipsis pattern. We have to store it on the list.
                var varData = scope[patternCur.car.value];
                var pair = new _pair3.default(codeNode.car, null);
                if (varData) {
                  varData.tail.cdr = pair;
                  varData.tail = pair;
                } else {
                  scope[patternCur.car.value] = {
                    type: LIST_WRAP,
                    head: pair,
                    tail: pair
                  };
                }
              }
            }
        } else if (patternCur.car.type === _value.PAIR) {
          if (codeNode.car == null) {
            var result = this.checkPattern(patternCur.car, new _pair3.default(), scope, patternEllipsis);
            if (!result) return false;
          } else {
            if (codeNode.car.type !== _value.PAIR) return false;
            var _result = this.checkPattern(patternCur.car, codeNode.car, scope, patternEllipsis);
            if (!_result) return false;
          }
        } else {
          // Nope
          throw new Error('Pattern datum is not supported yet. Please use ' + 'if (or when) in the template for now!');
        }
        codeNode = codeNode.cdr;
        codeLen--;
        if (patternEllipsis) {
          if (patternLen - 2 < codeLen) {
            // Ellipsis repeat.
            continue;
          }
        }
        patternCur = patternNext;
        patternNext = patternNext && patternNext.cdr;
        patternEllipsis = patternNext && patternNext.type === _value.PAIR && patternNext.car.type === _value.SYMBOL && patternNext.car.value === '...';
        patternLen--;
      }
      // Check CDR value too
      // TODO We should merge this with top node
      if (patternCur != null) {
        // Underflow
        if (patternCur.type === _value.SYMBOL) {
          if (codeNode == null) {
            // If codeNode is null - it failed (underflow).
            if (codeNode == null) return false;
          } else if (patternCur.value === '_') {
            // This always matches
          } else if (this.literals.indexOf(patternCur.value) !== -1) {
              // Literal identifier; check if code also matches.
              if (codeNode.type !== _value.SYMBOL || patternCur.value !== codeNode.value) return false;
            } else {
              // Pattern variable...
              // TODO check confliction
              if (!parentEllipsis) {
                // Flat pattern. We can just copy it to the scope
                scope[patternCur.value] = codeNode;
              } else {
                // Ellipsis pattern. We have to store it on the list.
                var _varData = scope[patternCur.value];
                var _pair = new _pair3.default(codeNode, null);
                if (_varData) {
                  _varData.tail.cdr = _pair;
                  _varData.tail = _pair;
                } else {
                  scope[patternCur.value] = {
                    type: LIST_WRAP,
                    head: _pair,
                    tail: _pair
                  };
                }
              }
            }
        } else if (patternCur.type === _value.PAIR) {
          if (codeNode == null) {
            var _result2 = this.checkPattern(patternCur, new _pair3.default(), scope, patternEllipsis);
            if (!_result2) return false;
          } else {
            if (codeNode.type !== _value.PAIR) return false;
            var _result3 = this.checkPattern(patternCur, codeNode, scope, patternEllipsis);
            if (!_result3) return false;
          }
        } else {
          // Nope
          throw new Error('Pattern datum is not supported yet. Please use ' + 'if (or when) in the template for now!');
        }
      } else if (codeNode != null) {
        // It failed (Overflow)
        return false;
      }
      return true;
    }
  }, {
    key: 'runTemplate',
    value: function runTemplate(template, scope) {
      var listLoc = arguments.length <= 2 || arguments[2] === undefined ? {} : arguments[2];

      if (template.type !== _value.PAIR) {
        // This is a special case, where only symbol or constant value can exist.
        if (template.type === _value.SYMBOL) {
          var scopeValue = scope[template.value];
          if (scopeValue) {
            if (scopeValue.type === LIST_WRAP) {
              // I think this won't get a list though.
              if (listLoc[template.value] == null) {
                listLoc[template.value] = scopeValue.head;
              }
              var listValue = listLoc[template.value];
              if (listValue === false) {
                // Underflow.
                return false;
              }
              listLoc[template.value] = listValue.cdr || false;
              return listValue.car;
            } else {
              return scopeValue;
            }
          } else {
            return template;
          }
        } else {
          // Just return template if this is a constant value.
          return template;
        }
      }
      // This doesn't support (... ...) producing single ellipsis!
      var outputHead = void 0,
          outputTail = void 0;
      var cur = template;
      var next = template.cdr;
      var ellipsis = next && next.type === _value.PAIR && next.car.type === _value.SYMBOL && next.car.value === '...';
      var cdrNode = false;
      function pushOutput(data) {
        if (cdrNode) {
          if (outputHead) {
            outputTail.cdr = data;
          } else {
            // This isn't really possible though.
            outputHead = outputTail = new _pair3.default(null, data);
          }
          return;
        }
        var pair = new _pair3.default(data);
        if (outputHead) {
          outputTail.cdr = pair;
          outputTail = pair;
        } else {
          outputHead = outputTail = pair;
        }
      }
      while (cur != null) {
        var current = void 0;
        if (cur.type === _value.PAIR) {
          current = cur.car;
        } else {
          current = cur;
          cdrNode = true;
        }
        if (current == null) {
          // Null value
          if (ellipsis) throw new Error('Unexpected ellipsis');
        } else if (current.type === _value.PAIR) {
          if (ellipsis) {
            // Run template until false comes out. Weird, huh?
            var listObj = Object.assign({}, listLoc);
            var output = this.runTemplate(current, scope, listObj);
            while (output !== false) {
              pushOutput(output);
              output = this.runTemplate(current, scope, listObj);
            }
          } else {
            var _output = this.runTemplate(current, scope, listLoc);
            if (_output === false) return false;
            pushOutput(_output);
          }
        } else if (current.type === _value.SYMBOL) {
          // Ignore ...
          if (current.value !== '...') {
            var _scopeValue = scope[current.value];
            if (ellipsis) {
              if (_scopeValue && _scopeValue.type === LIST_WRAP) {
                // Simply loop and copy all the values from the list.
                var node = _scopeValue.head;
                while (node != null) {
                  pushOutput(node.car);
                  node = node.cdr;
                }
              } else {
                // What??
                throw new Error(current.value + ' has unexpected ellipsis');
              }
            } else {
              if (_scopeValue) {
                if (_scopeValue.type === LIST_WRAP) {
                  if (listLoc[current.value] == null) {
                    listLoc[current.value] = _scopeValue.head;
                  }
                  var _listValue = listLoc[current.value];
                  if (_listValue === false || _listValue == null) {
                    // Underflow.
                    return false;
                  }
                  pushOutput(_listValue.car);
                  listLoc[current.value] = _listValue.cdr || false;
                } else {
                  pushOutput(_scopeValue);
                }
              } else {
                pushOutput(current);
              }
            }
          }
        } else {
          // Constant value
          if (ellipsis) throw new Error('Unexpected ellipsis');
          pushOutput(current);
        }
        cur = next;
        next = next && next.cdr;
        ellipsis = next && next.type === _value.PAIR && next.car.type === _value.SYMBOL && next.car.value === '...';
      }
      return outputHead || new _pair3.default();
    }
  }, {
    key: 'exec',
    value: function exec(code) {
      var node = this.rules;
      while (node != null) {
        var scope = {};
        if (this.checkPattern(node.car.car, code, scope)) {
          // Transform the code using template
          var result = this.runTemplate(node.car.cdr.car, scope);
          if (result === false) throw new Error('Data underflow');
          return result;
        }
        node = node.cdr;
      }
      return code;
    }
  }]);

  return SyntaxRules;
}();

exports.default = SyntaxRules;